#include <sys/stat.h>
#include <sys/conf.h>
#include <sys/mc_amd.h>
#include <sys/cpuvar.h>
#include <sys/modctl.h>
#include <sys/open.h>

#include "k10sensor.h"

#define DEV2INST(dev) (getminor(dev))

#define SENSOR_NAME "k10sensor"

#define K10_TERMTRIP 1
#define K10_HTC (1 << 1)
#define K10_SLEW (1 << 2)
#define DISABLED_VAL 0xffffffff
#define SENSOR_IDX(cpuNum, sensorNum)  (cpuNum ? 2 : 0) | ((sensorNum) ? 1 : 0)

#define K10_T_SENSOR_0 0
#define K10_T_SENSOR_1 1
#define K10_T_SENSOR_2 2
#define K10_T_SENSOR_3 3
#define K10_CPUS_SENSORS_TERMTRIP_HTC_SLEW 4
#define K10_DIODE_OFFSET  5
#define K10_T_HTC_HYSTERESIS 6
#define K10_T_HTC_MAX 7
#define K10_T_STC_HYSTERESIS 8
#define K10_T_STC_MAX 9
#define K10_T_CASE_MAX 10
#define K10_ALL 0xff

static int k10sensor_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int k10sensor_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int k10sensor_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd, void *arg,
	void **resultp);

static int k10sensor_open(dev_t *devp, int flag, int otyp, cred_t *cred);
static int k10sensor_close(dev_t dev, int flag, int otyp, cred_t *cred);
static int k10sensor_read(dev_t dev, struct uio *uiop, cred_t *credp);
static int k10sensor_ioctl(dev_t, int, intptr_t, int, cred_t *, int *);

static struct cb_ops k10sensor_cb_ops = {
	k10sensor_open,		/* open */
	k10sensor_close,	/* close */
	nodev,				/* strategy */
	nodev,				/* print */
	nodev,				/* dump */
	k10sensor_read,		/* read */
	nodev,				/* write */
	k10sensor_ioctl,	/* ioctl */
	nodev,				/* devmap */
	nodev,				/* mmap */
	nodev,				/* segmap */
	nochpoll,			/* chpoll */
	ddi_prop_op,		/* prop_op */
	NULL,				/* streamtab */
	D_MP | D_64BIT,		/* flags */
	CB_REV,				/* revision */
	nodev,				/* aread */
	nodev				/* awrite */
};

static struct dev_ops k10sensor_dev_ops = {
	DEVO_REV,					/* revision */
	0,							/* reference count */
	k10sensor_getinfo,			/* getinfo */
	nulldev,					/* obsolete identify */
	nulldev,					/* probe */
	k10sensor_attach,			/* attach */
	k10sensor_detach,			/* detach */
	nodev,						/* reset - unsupported */
	&k10sensor_cb_ops,			/* cb_ops */
	(struct bus_ops *) NULL,	/* bus operations entry points structure */
	nulldev						/* power */
#ifdef __SunOS_5_11
	,ddi_quiesce_not_needed		/* quiesce */
#endif
};

static struct modldrv md = {
	&mod_driverops,
	"AMD CPU temperature sensor",
	&k10sensor_dev_ops			/* dev_ops */
};

static struct modlinkage ml = {
	MODREV_1,
	&md,
	NULL
};

typedef struct k10sensor {
	/* info for the device this driver instance is attached to */
	dev_info_t *dip;
	/* collection of CPU chip related parameters, which are usually required to
	 * obtain other chip related infos */
	cpu_vars_t v[1];
	clock_t last_update;
	int t_ctrl;
	int t_ctrl_max;
	int t_diode_offset;
	int interval;
	int chip_id;
	kmutex_t lock[1];
} k10sensor_t;

static void *statep;

static int
getNumPhysCpus() {
	int count = 1;
	cpu_t *start = cpu_list;
	int last = cpuid_get_chipid(start);
	cpu_t *tcpu = start->cpu_next;
	for (;tcpu != start && tcpu != NULL; tcpu = tcpu->cpu_next); {
		int cid = cpuid_get_chipid(tcpu);
		if (cid != last) {
			count++;
			last = cid;
		}
	}
	return (count);
}

int
_init(void)
{
	int pcpus = getNumPhysCpus();
	cmn_err(CE_NOTE, "Inside _init %d cpus online", pcpus);
	ddi_soft_state_init(&statep, sizeof(k10sensor_t), pcpus);
	pcpus = mod_install(&ml);
	if (pcpus != 0) {
		cmn_err(CE_WARN, "_init: unable to allocate space for %d instances",
				pcpus);
		ddi_soft_state_fini(&statep);
	}
	return (pcpus);
}

int
_info(struct modinfo *modinfop)
{
	cmn_err(CE_NOTE, "Inside _info");
	return (mod_info(&ml, modinfop));
}

int
_fini(void)
{
	int err;
	cmn_err(CE_NOTE, "Inside _fini");
	err = mod_remove(&ml);
	if (err == 0) {
		ddi_soft_state_fini(&statep);
	}
	return (err);
}

static int
getChipId(k10sensor_t *k10p) {
	char *unitstr = NULL;
	long unitaddr;

	if (ddi_prop_lookup_string(DDI_DEV_T_ANY, k10p->dip, DDI_PROP_DONTPASS,
		"unit-address", &unitstr) != DDI_PROP_SUCCESS)
	{
		return (DDI_FAILURE);
	}
	if (ddi_strtol(unitstr, NULL, 16, &unitaddr) != 0) {
		unitaddr = 0;
	}
	ddi_prop_free(unitstr);
	k10p->chip_id = unitaddr - MC_AMD_DEV_OFFSET;
	return (k10p->chip_id < 0) ? (DDI_FAILURE) : (DDI_SUCCESS);
}

static int
getChipInfo(k10sensor_t *k10p) {
	struct cpuid_regs cp;
	cpu_t *start_cpu, *tmp_cpu;
	cpu_vars_t *v = k10p->v;
	/* once CPU list is initialized, it doesn't change its structure */
	tmp_cpu = start_cpu = cpu_list;
	do {
		int chip_id = cpuid_get_chipid(tmp_cpu);
		if (chip_id == k10p->chip_id) {
			start_cpu = NULL;
			break;
		}
		tmp_cpu = tmp_cpu->cpu_next;
	} while (tmp_cpu != start_cpu && tmp_cpu != NULL);
	if (start_cpu != NULL) {
		return -1;
	}
	cp.cp_eax = 1;
	cpuid_insn(tmp_cpu, &cp);
	v->xfamilyModel = AMD_XFAMILY_MODEL(cp.cp_eax);
	v->modelStepping = AMD_MODEL_STEPPING(cp.cp_eax);
	v->ebx = cp.cp_ebx;

	fillCategoryIndex(v);
	fillSiliconRevision(v);
	fillErratum(v);
	fillNBcapabilities(v);
	fillThermtrip(v);
	fillHtc(v);
	fillStc(v);
	return 0;
}

static int
destroyState(k10sensor_t *k10p, void *state) {
	if (k10p == NULL) {
		return (DDI_SUCCESS);
	}
	mutex_enter(k10p->lock);
	ddi_remove_minor_node(k10p->dip, NULL);
	if (k10p->v->pci_hdl) {
		pci_config_teardown(&k10p->v->pci_hdl);
	}
	ddi_set_driver_private(k10p->dip, NULL);
	mutex_exit(k10p->lock);
	mutex_destroy(k10p->lock);
	ddi_soft_state_free(state, ddi_get_instance(k10p->dip));
	return (DDI_SUCCESS);
}

static int
k10sensor_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	k10sensor_t *k10p;
	char buf[80];

	int inst = ddi_get_instance(dip);
	cmn_err(CE_NOTE, "Inside k10temp_attach #%d", inst);
	if (x86_vendor != X86_VENDOR_AMD ) {
		cmn_err(CE_NOTE, "No AMD CPU");
		return (DDI_FAILURE);
	}

	switch (cmd) {
		default:
			return (DDI_FAILURE);
		case DDI_RESUME:
			return (DDI_SUCCESS);
		case DDI_ATTACH:
			break;
	}

	if (ddi_soft_state_zalloc(statep, inst) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "attach: soft state zalloc failed on #%d\n", inst);
		return (DDI_FAILURE);
	}
	k10p = ddi_get_soft_state(statep, inst);
	ddi_set_driver_private(dip, k10p);
	k10p->dip = dip;
	if (getChipId(k10p) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "attach: getting chip ID failed on #%d", inst);
		goto failed;
	}
	/* unit-address = PCI_device_number,Function
	   which in this case is (0x18 + ChipId),3 */
	if (pci_config_setup(dip, &k10p->v->pci_hdl) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "attach: setup pci config failed on #%d", inst);
		goto failed;
	}
	(void) sprintf(buf, "%s-%d", SENSOR_NAME, inst);
	if (ddi_create_minor_node(dip, buf, S_IFCHR,
		inst, DDI_PSEUDO, 0) != DDI_SUCCESS)
	{
		cmn_err(CE_WARN, "attach: create minor node failed on #%d", inst);
		goto failed;
	}
	k10p->interval = drv_usectohz(200000); /* 200 ms */
	k10p->last_update = 0;
	mutex_init(k10p->lock, NULL, MUTEX_DRIVER, NULL);
	ddi_report_dev(dip);
	return (DDI_SUCCESS);

failed:
	destroyState(k10p, statep);
	return (DDI_FAILURE);
}

static int
k10sensor_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	k10sensor_t *k10p = ddi_get_driver_private(dip);
	cmn_err(CE_NOTE, "Inside k10temp_detach #%d cmd=%d", ddi_get_instance(dip),
			cmd);
	switch (cmd) {
		default:
			return DDI_FAILURE;
		case DDI_SUSPEND:
			return DDI_SUCCESS;
		case DDI_DETACH:
			break;
	}
	destroyState(k10p, statep);
	return DDI_SUCCESS;
}

static int
k10sensor_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd, void *arg, void **resultp)
{
	int res = DDI_FAILURE;
	dev_t dev;
	k10sensor_t *k10p;
	cmn_err(CE_NOTE, "Inside k10temp_getinfo");
	switch (cmd) {
		case DDI_INFO_DEVT2DEVINFO:
			dev = (dev_t)arg;
			k10p = ddi_get_soft_state(statep, DEV2INST(dev));
			if (k10p != NULL) {
				mutex_enter(k10p->lock);
				*resultp = k10p->dip;
				mutex_exit(k10p->lock);
				res = DDI_SUCCESS;
			} else {
				*resultp = NULL;
			}
			break;
		case DDI_INFO_DEVT2INSTANCE:
			dev = (dev_t)arg;
			*resultp = (void *)(uintptr_t) DEV2INST(dev);
			res = DDI_SUCCESS;
			break;
		default:
			break;
	}
	return res;
}

	/* ddi_get_lbolt(void); */

static int
k10sensor_open(dev_t *devp, int flag, int otyp, cred_t *cred)
{
	int instance = getminor(*devp);
	cmn_err(CE_NOTE, "Inside k10temp_open #%d", instance);
	return otyp != OTYP_CHR ? EINVAL : DDI_SUCCESS;
}

static int
k10sensor_close(dev_t dev, int flag, int otyp, cred_t *cred)
{
	int instance = getminor(dev);
	cmn_err(CE_NOTE, "Inside k10temp_close #%d", instance);
	return otyp != OTYP_CHR ? EINVAL : DDI_SUCCESS;
}

static int
k10sensor_read(dev_t dev, struct uio *uiop, cred_t *credp)
{
	char buf[4096];
	int instance = getminor(dev);
	size_t blen, len = uiop->uio_resid;

	cmn_err(CE_NOTE, "Inside k10temp_read #%d", instance);

	if (len == 0) {
		return (0);
	}
	/* prepare stats */
	sprintf(buf, "Temperature: %d\n", 0);
	blen = strlen(buf);
	if (uiop->uio_offset < 0 || blen < uiop->uio_offset) {
		return EINVAL;
	}
	if (len > blen - uiop->uio_offset) {
		len = blen - uiop->uio_offset;
	}
	if (len == 0) {
		return 0;
	}
	return (uiomove((void *)(buf + uiop->uio_offset), len, UIO_READ, uiop));
}

static int
k10sensor_ioctl(dev_t dev, int cmd, intptr_t arg, int mode, cred_t *credp,
	int *rvalp)
{
	int instance = getminor(dev);
	cmn_err(CE_NOTE, "Inside k10temp_ioctl #%d", instance);
	switch (cmd) {
		default:
			return ENOTTY;
		case K10IOCLMT:
			if (! (mode & FREAD)) {
				return EACCES;
			}
			break;
	}
	/* re-read limits */
	return DDI_SUCCESS;
}
