/** 
 * $Id$
 *
 * * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at
 * http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each file.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * All Rights reserved.
 * CDDL HEADER END
 */
/*
 * Copyright (c) 2011,2012 by Jens Elkner,
 * Otto-von-Guericke Universitaet Magdeburg. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include "k10sensor.h"

#define AMD_VENDOR_ID 0x1022
#define AMD_0Fh_DRAM_CONTROL_DEV_ID 0x1102 /* Athlon64/Opteron (early K8) */
#define AMD_0Fh_MISC_CONTROL_DEV_ID 0x1103 /* Athlon64/Opteron (early K8) */
#define AMD_10h_DRAM_CONTROL_DEV_ID 0x1202 /* K10 */
#define AMD_10h_MISC_CONTROL_DEV_ID 0x1203 /* K10 */
#define AMD_11h_DRAM_CONTROL_DEV_ID 0x1302 /* Mobiles? */
#define AMD_11h_MISC_CONTROL_DEV_ID 0x1303 /* Mobiles? */
#define AMD_12h_14h_DRAM_CONTROL_DEV_ID 0x1702 /* Llano, Brazos (C/E/G) */
#define AMD_12h_14h_MISC_CONTROL_DEV_ID 0x1703 /* Llano, Brazos (C/E/G) */
#define AMD_15h_DRAM_CONTROL_DEV_ID 0x1602	/* Bulldozer */
#define AMD_15h_MISC_CONTROL_DEV_ID 0x1603	/* Bulldozer */

static const uint32_t knownDevId[] = {
	AMD_0Fh_MISC_CONTROL_DEV_ID,
	AMD_10h_MISC_CONTROL_DEV_ID,
	AMD_11h_MISC_CONTROL_DEV_ID,
	AMD_12h_14h_MISC_CONTROL_DEV_ID,
	AMD_15h_MISC_CONTROL_DEV_ID
};

static boolean_t
getCPUid(uint32_t *eax, uint32_t *ebx) {
	uint32_t regs[4];
	int device = open("/dev/cpu/self/cpuid", O_RDONLY);
	if (device == -1) {
		return B_FALSE;
	}
	if (pread(device, regs, sizeof (regs), 0x80000001) != sizeof (regs)) {
		(void) close(device);
		return B_FALSE;
	}
	(void) close(device);
	*eax = regs[0];
	*ebx = regs[1];
	return B_TRUE;
}

#define ERRATUM_MSG \
"WARNING: AMD %s erratum #%d (inaccurrate sensor data) applies!\n" \
"         DON'T USE the values read for anything but fun!\n"

#define FORMAT "%24s:\t%s\n"

#define PRINT_YORN(what, val) \
	fprintf(stdout, FORMAT, what, (val) ? "yes" : "no")

#define PRINT_ENABLED(what, val)  \
	fprintf(stdout, FORMAT, what,  \
		((val) == INT32_MIN || (val) == INT32_MAX) \
			? "n/a" \
			: ((val) != 0 ? "enabled" : "disabled"))

#define PRINT_VAL(what, val) \
	{ \
		char buf[80]; \
		buf[0] = '\0'; \
		if ((val) == INT32_MIN || (val) == INT32_MAX) { \
			strcat(buf, "n/a"); \
		} else { \
			sprintf(buf, "%5.1f", (val)/1000.0 ); \
		} \
		fprintf(stdout, FORMAT, what, buf); \
	}
#define PRINT_SEPARATOR fprintf(stdout, "\n");

int main(int argc, char** argv)
{
	struct pci_id_match match;
	struct pci_device_iterator *iter;
	cpu_vars_t cv[1];
	uint32_t eax, ebx;
	int32_t stcIdx = 0;
	const char *STC_NAMES[][4] = {
		{ "STC", "STC Limit", "STC Hysteresis", "STC slew-controlled" },
		{ "LHTC", "LHTC Limit", "LHTC Hysteresis", "LHTC slew-controlled" }
	};

	/* since we know, that a box can't have mixed CPU families, we do
	   not cycle through all knownDevIds but obtain the info directly */
	if (!getCPUid(&eax, &ebx)) {
		perror("Couldn't obtain CPU ID infos");
		exit(EXIT_FAILURE);
	}
	bzero(cv, sizeof(cv));
	cv->xfamilyModel = AMD_XFAMILY_MODEL(eax);
	cv->modelStepping = AMD_MODEL_STEPPING(eax);
	cv->ebx = ebx;
	fillCategoryIndex(cv);
	/* Actually some different silicon revision CPUs can be mixed in a box
	 * (see Revision Guide if in doubt).
	 * However, to keep this easy, we assume here, all have the same. The
     * kernel driver handles each CPU separately - provides the final word. */
	fillSiliconRevision(cv);
	fillErratum(cv);
	stcIdx = (cv->xfamilyModel == 0x30 || cv->xfamilyModel == 0x50) ? 1 : 0;

	if (cv->xfamilyModel < 0x04) {
		match.device_id = AMD_0Fh_MISC_CONTROL_DEV_ID;
	} else if (cv->xfamilyModel < 0x10) {
		match.device_id = AMD_0Fh_MISC_CONTROL_DEV_ID;
	} else if (cv->xfamilyModel < 0x20) {
		match.device_id = AMD_10h_MISC_CONTROL_DEV_ID;
	} else if (cv->xfamilyModel < 0x30) {
		match.device_id = AMD_11h_MISC_CONTROL_DEV_ID;
	} else if (cv->xfamilyModel < 0x60) {
		match.device_id = AMD_12h_14h_MISC_CONTROL_DEV_ID;
	} else {
		match.device_id = AMD_15h_MISC_CONTROL_DEV_ID;
	}

	if (cv->erratum) {
		fprintf(stdout, ERRATUM_MSG, cat_range_map[cv->catIdx].name,
			cv->erratum);
	}
	fprintf(stdout,	"\nSilicon CPU Revision: %s (%02x-%02x)\n"
					"Category:             %s\n",
		cv->revision, cv->xfamilyModel, cv->modelStepping,
		cat_range_map[cv->catIdx].name);

	if (pci_system_init()) {
		perror("Couldn't initialize PCI system");
		exit(EXIT_FAILURE);
	}
	match.vendor_id = AMD_VENDOR_ID;
	match.subdevice_id = PCI_MATCH_ANY;
	match.subvendor_id = PCI_MATCH_ANY;
	match.device_class = 0; /* PCI_CLASS_BRIDGE, but IDs are sufficient ;-) */
	match.device_class_mask = 0;

	iter = pci_id_match_iterator_create(&match);
	while ((cv->pci_hdl = pci_device_next(iter)) != NULL) {
		int htcCapable, val, slewed;
		fillThermtrip(cv);
		fillNBcapabilities(cv);
		fillHtc(cv);
		fillStc(cv);
		fprintf(stdout, "\n"
			"#####################################\n"
			"CPU #%02d  %s\n"
			"#####################################\n",
			cv->pci_hdl->dev, (cv->flags & DC_0Fh ? "DualCore" : ""));

		htcCapable = cv->flags & HTC_CAPABLE;
		PRINT_ENABLED("THERMTRIP", cv->flags & HTC_THERMTRIP);
		PRINT_YORN("HTC capable", htcCapable);
		PRINT_SEPARATOR
		PRINT_VAL("T Case/Die Max", cv->t_case_max);
		PRINT_VAL("Diode Offset", cv->diode_offset);
		PRINT_SEPARATOR
		PRINT_ENABLED("HTC", cv->limits[HTC].flags & TC_ENABLED);
		PRINT_YORN("HTC slew-controlled", cv->limits[HTC].flags & TC_SLEWED);
		PRINT_VAL("HTC Limit", cv->limits[HTC].tcrit);
		PRINT_VAL("HTC Hysteresis", cv->limits[HTC].threshold);
		PRINT_SEPARATOR
		PRINT_ENABLED(STC_NAMES[stcIdx][0], cv->limits[STC].flags & TC_ENABLED);
		PRINT_YORN(STC_NAMES[stcIdx][3], cv->limits[STC].flags & TC_SLEWED);
		PRINT_VAL(STC_NAMES[stcIdx][1], cv->limits[STC].tcrit);
		PRINT_VAL(STC_NAMES[stcIdx][2], cv->limits[STC].threshold);
		PRINT_SEPARATOR
		if (cv->xfamilyModel < 0x04) {
			continue; /* no control temp available */
		}
		if (cv->xfamilyModel < 0x10 && (cv->flags & DC_0Fh)) {
			/* 0Fh DualCore */
			val = getControlTemp(cv, B_FALSE);
			PRINT_VAL("Control Temp (C0)", val - cv->diode_offset);
			val = getControlTemp(cv, B_TRUE);
			PRINT_VAL("Control Temp (C1)", val - cv->diode_offset);
		} else {
			/* 10h + or 0Fh single core */
			val = getControlTemp(cv, 0);
			PRINT_VAL("Control Temp", val - cv->diode_offset);
		}
	}
	pci_iterator_destroy(iter);
	pci_system_cleanup();
	return (EXIT_SUCCESS);
}

/* vim: tabstop=4:shiftwidth=4:noexpandtab
 */
